unit hat_eeprom_ctlr;

interface

uses
  {$IFDEF UNIX}
  FileUtil, BaseUnix,
  {$ENDIF}
  Classes, SysUtils, crt;

type
  // Eeprom size and buffer
  TEpromCapacity = (ec16, ec32, ec64, ec128, ec256, ec512);
  TEpromBuffer = array of byte;

  // Address + value bloc for byte manipulation
  TAddrValue = record
    case dummy: boolean of
      True: (bytes: array[0..2] of byte);
      False: (Addr: UInt16; Value: byte);
  end;

  PByte = ^byte;

  TheHeader = record
    signature: array[0..3] of char;   // = "R-Pi"
    version: byte;                    // = 0x01
    reserved: byte;                   // = 0x00
    numAtoms: UInt16;                 // total atoms in EEPROM
    eepLen : UInt32;                  // total length in bytes of all eeprom data (including this header)
  end;

  TheaVendor = record
    aType: UInt16;        // 0x0001 = vendor info
    count: UInt16;        // incrementing atom count
    dLen: UInt32;         // length in bytes of data+CRC
    UUID: TGuid;
    pid: UInt16;          // product ID
    pver: UInt16;         // product version
    vslen: byte;          // vendor AnsiString length (bytes)
    vplen: byte;          // product AnsiString length (bytes)
    vstr: AnsiString;     // ASCII vendor AnsiString e.g. "ACME Technology Company"
    pstr: AnsiString;     // ASCII product AnsiString e.g. "Special Sensor Board"
    crc16: UInt16;
  end;

  TheaGPIO = record
    aType: UInt16;        // 0x0002 = GPIO map
    count: UInt16;        // incrementing atom count
    dLen: UInt32;         // length in bytes of data+CRC
    bank_drive: byte;     // bank drive strength/slew/hysteresis, BCM2835 can only set per bank, not per IO
                             // Bits in byte:
                             // [3:0] drive       0=leave at default, 1-8=drive*2mA, 9-15=reserved
                             // [5:4] slew        0=leave at default, 1=slew rate limiting, 2=no slew limiting, 3=reserved
                             // [7:6] hysteresis  0=leave at default, 1=hysteresis disabled, 2=hysteresis enabled, 3=reserved
    power: byte;          // [1:0] back_power  0=board does not back power Pi
                             // 1=board back powers and can supply up to 1.3A to the Pi
                             // 2=board back powers and can supply up to 2A to the Pi
                             // 3=reserved
                             // If back_power=2 high current USB mode is automatically enabled.
                          // [7:2] reserved    set to 0
    IO: array[0..27] of byte;
                          //  1 byte per IO pin, bits in each byte:
                             //  [2:0] func_sel    GPIO function as per FSEL GPIO register field in BCM2835 datasheet
                             //  [4:3] reserved    set to 0
                             //  [6:5] pulltype    0=leave at default setting,  1=pullup, 2=pulldown, 3=no pull
                             //  [  7] is_used     1=board uses this pin, 0=not connected and therefore not used
    crc16: UInt16;
  end;

  TheaCustom = record
    aType: UInt16;        // 0x0004 = manufacturer custom data
    count: UInt16;        // incrementing atom count
    dLen: UInt32;         // length in bytes of data+CRC
    data: array of byte;  // N bytes, N = dlen-2
    crc16: UInt16;
    dataOffset: UInt32;   // to save the position
  end;

  TAtoms = record
    Header: TheHeader;
    Vendor: TheaVendor;
    GPIO: TheaGPIO;
    DeviceTreeBlob: TheaCustom;
    Customs: array of TheaCustom;
  end;



  { THat_eeprom_ctlr }
  THat_eeprom_ctlr = class(TComponent)
  private
    FAtoms: TAtoms;
    FHandle: Cint;
    FInitialized: boolean;
    FLastError: AnsiString;
    FeType: TEpromCapacity;
    FeBuf: TEpromBuffer;
    bufAddr: TAddrValue;

    procedure Initialize;
    function CRC16CCITT(PBytes: PByte; ALen: UInt16): UInt16;
    procedure SeteType(AValue: TEpromCapacity);

    procedure ParseGPIO(ACount: UInt16; AdLen: UInt32; ABuffer: TEpromBuffer;
                        var Offset: UInt32);
    procedure ParseVendor(ACount: UInt16; AdLen: UInt32; ABuffer: TEpromBuffer;
                          var Offset: UInt32);
    procedure ParseCustom(ACount: UInt16; AdLen: UInt32; ABuffer: TEpromBuffer;
                         var Offset: UInt32);

  public
    // Eeprom type and content
    property EpromType: TEpromCapacity read FeType write SeteType;
    property EpromBuffer: TEpromBuffer read FeBuf write FeBuf;

    property Atoms: TAtoms read FAtoms write FAtoms;

    property Initialized: boolean read FInitialized write FInitialized;

    // error handling
    property LastError : AnsiString read FLastError;

    constructor Create(AOwner: TComponent; AType: TEpromCapacity);
    destructor Destroy; override;

    // basic read / write
    function ReadByte(Addr: UInt16): byte;
    procedure WriteByte(Addr: UInt16; Data: byte);
    procedure ReadArray(Addr: UInt16; var ByteArray: TEpromBuffer); overload;
    procedure ReadArray(Addr: UInt16; var ByteArray: TEpromBuffer; ALength: UInt16); overload;
    procedure WriteArray(Addr: UInt16; ByteArray: TEpromBuffer); overload;
    procedure WriteArray(Addr: UInt16; ByteArray: TEpromBuffer; ALength: UInt16); overload;

    // eeprom transfer
    procedure EpromToBuffer;
    procedure BufferToEprom;   // flash current buffer
    procedure HexFileToBuffer(AFilename: AnsiString);
    procedure BufferToHexFile(AFilename: AnsiString);

    procedure BlankEprom;

    // atoms operations
    procedure ParseBuffer;
    procedure NewUUID;
    procedure ResetAtoms;
    procedure InitAtoms(sVendor, sProduct: AnsiString; PID, VersionNo : UInt16);
    function AddCustom(const AData; ALen:UInt16): UInt16;
    procedure AtomsToBuffer;

    // user data : save and retrieve (previously initialized)
    procedure SaveToCustom(AtomIx: UInt16; const AData; ALen:UInt16);
    procedure GetFromCustom(AtomIx: UInt16; var AData; ALen:UInt16);

   end;


implementation

const
   RPi_I2C0_DevPath=      '/dev/i2c-0';
   Eeprom_addr: Cint=     $50;
   I2C_SLAVE: Cint=       1795;
   //Eeprom capacity, page size,
   eSize: array[TEpromCapacity] of UInt32 = (2048, 4096, 8192, 16384, 32768, 65536);
   PageSize = 16;         // 16 bytes per row (can be larger depend on eeprom manufacturer)
   PageMask: UInt16 = (PageSize-1);
   AddrSize = 2;          //  2 bytes for address
   WriteTime = 10;        // 10 ms

   CR= #10;
   LF= #13;
   {$IFDEF UNIX}
     EOL = CR;
   {$ELSE no UNIX}
     EOL = CR + LF;
   {$ENDIF}

   AtomStartLen = SizeOf(TheaVendor.aType) + SizeOf(TheaVendor.count)
                   + SizeOf(TheaVendor.dLen);

{ THat_eeprom_ctlr }
{==============================================================================}

constructor THat_eeprom_ctlr.Create(AOwner: TComponent; AType: TEpromCapacity);
begin
  inherited Create(AOwner);
  try
    FLastError:= '';
    FHandle:= fpopen(RPi_I2C0_DevPath, O_RDWR);
    fpIOCtl(FHandle, I2C_SLAVE, Pointer(Eeprom_addr));
    FLastError:= '';
    FeType:= AType;
    SetLength(FeBuf, eSize[AType]);
    Initialize;
  except
    FLastError:= 'Error create hat eeprom controler';
  end;
end;

procedure THat_eeprom_ctlr.Initialize;
begin
  try
    EpromToBuffer;
    ParseBuffer;
    FInitialized:= True;
  except
    FInitialized:= False;
    FLastError:= 'Error initialization';
  end;
end;


destructor THat_eeprom_ctlr.Destroy;
begin
  try
    fpclose(FHandle);
  finally
    inherited;
  end;
end;

procedure THat_eeprom_ctlr.SeteType(AValue: TEpromCapacity);
begin
  if FeType=AValue then Exit;
  FeType:=AValue;
  SetLength(FeBuf, eSize[AValue]);
end;

function THat_eeprom_ctlr.ReadByte(Addr: UInt16): byte;
begin
  try
    FLastError:= '';
    bufAddr.bytes[1]:= byte(Addr and $FF);
    bufAddr.bytes[0]:= byte(Addr shr 8);
    fpwrite(FHandle, bufAddr.bytes[0], 2);
    fpread(FHandle, bufAddr.Value, 1);
  except
    FLastError:= 'Error read value';
    Exit;
  end;
  Result:= bufAddr.Value;
end;

procedure THat_eeprom_ctlr.WriteByte(Addr: UInt16; Data: byte);
begin
  try
    FLastError:= '';
    bufAddr.bytes[1]:= byte(Addr and $FF);
    bufAddr.bytes[0]:= byte(Addr shr 8);
    bufAddr.Value:= Data;
    fpwrite(FHandle, bufAddr.bytes[0], 3);
    Delay(WriteTime); // 10 ms delay
  except
    FLastError:= 'Error write value';
  end;
end;

procedure THat_eeprom_ctlr.ReadArray(Addr: UInt16; var ByteArray: TEpromBuffer);
begin
  ReadArray(Addr, ByteArray, Length(ByteArray)-Addr);
end;

procedure THat_eeprom_ctlr.ReadArray(Addr: UInt16; var ByteArray: TEpromBuffer; ALength: UInt16);
begin
  SetLength(ByteArray, ALength);
  try
    FLastError:= '';
    bufAddr.bytes[1]:= byte(Addr and $FF);
    bufAddr.bytes[0]:= byte(Addr shr 8);
    fpwrite(FHandle, bufAddr.bytes[0], 2);
    fpread(FHandle, ByteArray[0], ALength);
  except
    FLastError:= 'Error read array';
    Exit;
  end;
end;

procedure THat_eeprom_ctlr.WriteArray(Addr: UInt16; ByteArray: TEpromBuffer);
begin
  WriteArray(Addr, ByteArray, Length(ByteArray)-Addr);
end;

procedure THat_eeprom_ctlr.WriteArray(Addr: UInt16; ByteArray: TEpromBuffer; ALength: UInt16);
const
   BufferLength = AddrSize + PageSize;
var iAddr, Offset, dLength: UInt16;
    wBuffer: array [0..BufferLength-1] of byte;
begin
  try
    FLastError:= '';
    iAddr:= Addr;
    repeat
      wBuffer[1]:= byte(iAddr and $FF);
      wBuffer[0]:= byte(iAddr shr 8);
      if iAddr and PageMask = 0
        then dLength:= PageSize
        else dLength:= PageSize - (iAddr and PageMask);
      if iAddr + dLength > Addr + ALength
        then dLength:= Addr + ALength - iAddr;
      for Offset:= 0 to dLength-1 do wBuffer[AddrSize + Offset]:= ByteArray[iAddr + Offset];
      fpwrite(FHandle, wBuffer, AddrSize + dLength);
      Delay(WriteTime);
      iAddr:= iAddr + dLength;
    until iAddr >= Addr+ALength;
  except
    FLastError:= 'Error write array';
  end;
end;

procedure THat_eeprom_ctlr.EpromToBuffer;
begin
  ReadArray(0, FeBuf);
end;

procedure THat_eeprom_ctlr.BufferToHexFile(AFilename: AnsiString);
const sHex=  ':10%s00%s';
      sHexEOF=  ':00000001FF' + EOL;
var OutList: TStringList;
    Addr, StartAddr: UInt16;
    CS: byte;
    HexData, OutStr: AnsiString;
begin
  try
   FLastError:= '';
   try
    OutList:= TStringList.Create;
    OutList.LineBreak := EOL;

    HexData:= '';
    CS:= $10;
    for Addr:= 0 to Length(FeBuf)-1 do
    begin
      HexData:= HexData + IntToHex(FeBuf[Addr], 2);
      CS:= CS + FeBuf[Addr];
      if (Addr and $F = $F)
      then begin
        StartAddr:= Addr and $FFF0;
        CS:= not( CS + byte(StartAddr shr 8) + byte(StartAddr and $F0) ) + 1;
        OutStr:= Format(sHex, [IntToHex(StartAddr, 4), HexData])
                         + IntToHex(CS, 2);
        OutList.Add(OutStr);
        HexData:= '';
        CS:= $10;
      end;
    end;
    OutList.Add(sHexEOF);
    OutList.SaveToFile(AFilename);
   except
    FLastError:= 'Error write hex file';
   end;
  finally
    OutList.Free;
  end;
 end;

procedure THat_eeprom_ctlr.BufferToEprom;
begin
  WriteArray(0, FeBuf);
end;

procedure THat_eeprom_ctlr.HexFileToBuffer(AFilename: AnsiString);
const
   rcData = 0;
   rcEOF  = 1;
var InList: TStringList;
    i, li, StartAddr, Addr, DataCount: UInt16;
    RecordType: byte;
    CurLine: AnsiString;

  function HexCharToByte(HexChar: char): byte;
  begin
    if HexChar <= '9'
      then Result:= byte(HexChar) - byte('0')
      else Result:= byte(HexChar) + 10 - byte('A');
  end;

  function HexToByte(HexStr: AnsiString): byte;
  begin
    Result:= HexCharToByte(HexStr[1]) shl 4 + HexCharToByte(HexStr[2]);
  end;

begin
  try
   FLastError:= '';
   try
    InList:= TStringList.Create;
    InList.LineBreak := EOL;
    InList.LoadFromFile(AFilename);

    for li:= 0 to InList.Count-1 do
    begin
      CurLine:= InList.Strings[li];
      if (CurLine <> '') and (CurLine[1] = ':') then
      begin
        RecordType:= HexToByte(Copy(InList[li], 8, 2));
        if RecordType = rcEOF then Break;
        if RecordType = rcData then
        begin
          DataCount:= HexToByte(Copy(CurLine, 2, 2));
          StartAddr:= HexToByte(Copy(CurLine, 4, 2)) shl 8
                              + HexToByte(Copy(CurLine, 6, 2)) ;
          for i:= 0 to DataCount - 1 do
          begin
            Addr:= StartAddr + i;
            if Addr < Length(FeBuf)
              then FeBuf[Addr]:= byte(HexToByte(Copy(CurLine, 10 + 2*i, 2)));
          end;
        end;
      end;
    end;

   except
    FLastError:= 'Error read hex file';
   end;
  finally
    InList.Free;
  end;
end;

procedure THat_eeprom_ctlr.BlankEprom;
const
   BufferLength = AddrSize + PageSize;
var i, Addr: UInt16;
    wBuffer: array [0..BufferLength-1] of byte;
begin
  try
    FLastError:= '';
    for i:= AddrSize to BufferLength-1 do wBuffer[i]:= $FF;
    Addr:= 0;
    repeat
      wBuffer[1]:= byte(Addr and $FF);
      wBuffer[0]:= byte(Addr shr 8);
      fpwrite(FHandle, wBuffer, BufferLength);
      Delay(WriteTime);
      Addr:= Addr + PageSize;
    until Addr >= Length(FeBuf);
  except
    FLastError:= 'Error blank eeprom';
  end;
end;


/////////////////   BUFFER OPERATIONS    ///////////////////////////////////////
procedure THat_eeprom_ctlr.ResetAtoms;
// Init Header + Vendor
begin
  // Header
  FAtoms.Header.numAtoms:= 0;
  FAtoms.Header.eepLen:= 0;
  // Vendor
  FAtoms.Vendor.count:= 0;
  FAtoms.Vendor.dLen:= 0;
  // GPIO
  FAtoms.GPIO.count:= 0;
  FAtoms.GPIO.dLen:= 0;
  // Tree
  FAtoms.DeviceTreeBlob.count:= 0;
  FAtoms.DeviceTreeBlob.dLen:= 0;
  //Custom
  SetLength(FAtoms.Customs, 0);
end;

procedure THat_eeprom_ctlr.ParseBuffer;
// set the atoms from content
var Offset, pLen, dLen: UInt32;
    i, aType, Count: UInt16;
begin
  ResetAtoms;
  try
    FLastError:= '';
    // check signature
    offset:= 0;
    pLen:= 4;
    Move(FeBuf[0], FAtoms.Header.signature, pLen);
    if FAtoms.Header.signature <> 'R-Pi'
    then begin
      FAtoms.Header.numAtoms:= 0;
      FAtoms.Header.eepLen:= 0;
      Exit;
    end;
    // set Header
    Offset:= pLen;
    FAtoms.Header.version:= byte((@FeBuf[Offset])^);
    Offset:= Offset + SizeOf(FAtoms.Header.version);
    FAtoms.Header.reserved:= byte((@FeBuf[Offset])^);
    Offset:= Offset + SizeOf(FAtoms.Header.reserved);
    FAtoms.Header.numAtoms:= UInt16((@FeBuf[Offset])^);
    Offset:= Offset + SizeOf(FAtoms.Header.numAtoms);
    FAtoms.Header.eepLen:= UInt32((@FeBuf[Offset])^);
    Offset:= Offset + SizeOf(FAtoms.Header.eepLen);
    // other atoms (start is common)
    for i:= 1 to FAtoms.Header.numAtoms do
    begin
      aType:= UInt16((@FeBuf[Offset])^);
      Offset:= Offset + SizeOf(FAtoms.Vendor.aType);
      Count:= UInt16((@FeBuf[Offset])^);
      Offset:= Offset + SizeOf(FAtoms.Vendor.Count);
      dLen:= UInt32((@FeBuf[Offset])^);
      Offset:= Offset + SizeOf(FAtoms.Vendor.dLen);
      case aType of
        1 : ParseVendor(Count, dLen, FeBuf, Offset);
        2 : ParseGPIO(Count, dLen, FeBuf, Offset);
        //3 : ParseTree();    // not implemented
        4 : ParseCustom(Count, dLen, FeBuf, Offset);
      end;
    end;
  except
    FLastError:= 'Error parsing datas';
  end;

end;

procedure THat_eeprom_ctlr.ParseVendor(ACount: UInt16; AdLen: UInt32; ABuffer: TEpromBuffer; var Offset: UInt32);
// set the atoms from content
begin
  FAtoms.Vendor.aType:= 1;
  FAtoms.Vendor.count:= ACount;
  FAtoms.Vendor.dLen:= AdLen;

  FAtoms.Vendor.UUID:= TGuid((@ABuffer[Offset])^);
  Offset:= Offset + SizeOf(FAtoms.Vendor.UUID);
  FAtoms.Vendor.pid:= UInt16((@ABuffer[Offset])^);   // product ID
  Offset:= Offset + SizeOf(FAtoms.Vendor.pid);
  FAtoms.Vendor.pver:= UInt16((@ABuffer[Offset])^);  // product version
  Offset:= Offset + SizeOf(FAtoms.Vendor.pver);
  FAtoms.Vendor.vslen:= byte((@ABuffer[Offset])^);   // vendor AnsiString length
  Offset:= Offset + SizeOf(FAtoms.Vendor.vslen);
  FAtoms.Vendor.vplen:= byte((@ABuffer[Offset])^);   // product AnsiString length
  Offset:= Offset + SizeOf(FAtoms.Vendor.vplen);
  SetString(FAtoms.Vendor.vstr, @ABuffer[Offset], FAtoms.Vendor.vslen);
  Offset:= Offset + FAtoms.Vendor.vslen;
  SetString(FAtoms.Vendor.pstr, @ABuffer[Offset], FAtoms.Vendor.vplen);
  Offset:= Offset + FAtoms.Vendor.vplen;

  FAtoms.Vendor.crc16:= UInt16((@ABuffer[Offset])^);
  Offset:= Offset + SizeOf(FAtoms.Vendor.crc16);
end;

procedure THat_eeprom_ctlr.ParseGPIO(ACount: UInt16; AdLen: UInt32; ABuffer: TEpromBuffer; var Offset: UInt32);
// set the GPIO atom
begin
  FAtoms.GPIO.aType:= 2;
  FAtoms.GPIO.count:= ACount;
  FAtoms.GPIO.dLen:= AdLen;

  FAtoms.GPIO.bank_drive:= byte((@ABuffer[Offset])^);
  Offset:= Offset + SizeOf(FAtoms.GPIO.bank_drive);
  FAtoms.GPIO.power:= byte((@ABuffer[Offset])^);
  Offset:= Offset + SizeOf(FAtoms.GPIO.power);
  Move(ABuffer[Offset], FAtoms.GPIO.IO[0], Length(FAtoms.GPIO.IO));
  Offset:= Offset + Length(FAtoms.GPIO.IO);

  FAtoms.GPIO.crc16:= UInt16((@ABuffer[Offset])^);
  Offset:= Offset + SizeOf(FAtoms.GPIO.crc16);
end;

procedure THat_eeprom_ctlr.ParseCustom(ACount: UInt16; AdLen: UInt32; ABuffer: TEpromBuffer; var Offset: UInt32);
// set the atoms from content
var iCustom, cLen: UInt16;
begin
  iCustom:= Length(FAtoms.Customs);
  SetLength(FAtoms.Customs, iCustom+1);
  FAtoms.Customs[iCustom].aType:= 3;
  FAtoms.Customs[iCustom].count:= ACount;
  FAtoms.Customs[iCustom].dLen:= AdLen;
  FAtoms.Customs[iCustom].dataOffset:= Offset;

  cLen:= AdLen - SizeOf(FAtoms.Customs[iCustom].crc16);
  SetLength(FAtoms.Customs[iCustom].data, cLen);
  Move(ABuffer[Offset], FAtoms.Customs[iCustom].data[0], cLen);
  Offset:= Offset + cLen;

  FAtoms.Customs[iCustom].crc16:= UInt16((@ABuffer[Offset])^);
  Offset:= Offset + SizeOf(FAtoms.Customs[iCustom].crc16);
end;


procedure THat_eeprom_ctlr.NewUUID;
// generate new UUID
begin
  CreateGUID(FAtoms.Vendor.UUID);
end;

procedure THat_eeprom_ctlr.InitAtoms(sVendor, sProduct: AnsiString; PID, VersionNo : UInt16);
// Init Header + Vendor
var i: integer;
begin
  ResetAtoms;
  // Header
  FAtoms.Header.signature:= 'R-Pi';
  FAtoms.Header.version:= 1;
  FAtoms.Header.reserved:= 0;
  FAtoms.Header.numAtoms:=2;
//  FAtoms.Header.eepLen:= SizeOf(FAtoms.Header);
  // Vendor
  FAtoms.Vendor.aType:= 1;
  FAtoms.Vendor.count:= 0;
  NewUUID;
  FAtoms.Vendor.pid:= PID;
  FAtoms.Vendor.pver:= VersionNo;
  FAtoms.Vendor.vslen:= length(sVendor);
  FAtoms.Vendor.vplen:= length(sProduct);
  FAtoms.Vendor.vstr:= sVendor;
  FAtoms.Vendor.pstr:= sProduct;
  FAtoms.Vendor.dLen:= SizeOf(FAtoms.Vendor.UUID)
                    + SizeOf(FAtoms.Vendor.pid) + SizeOf(FAtoms.Vendor.pid)
                    + SizeOf(FAtoms.Vendor.vslen) + SizeOf(FAtoms.Vendor.vplen)
                    + FAtoms.Vendor.vslen + FAtoms.Vendor.vplen
                    + SizeOf(FAtoms.Vendor.crc16);
  FAtoms.Vendor.crc16:= CRC16CCITT(@FAtoms.Vendor, AtomStartLen+FAtoms.Vendor.dLen-SizeOf(FAtoms.Vendor.crc16));
  // GPIO
  FAtoms.GPIO.aType:= 2;
  FAtoms.GPIO.count:= 1;
  FAtoms.GPIO.dlen:= SizeOf(FAtoms.GPIO)-AtomStartLen;
  FAtoms.GPIO.bank_drive:= 0;
  FAtoms.GPIO.power:= 0;
  for i:= 0 to Length(FAtoms.GPIO.IO)-1 do FAtoms.GPIO.IO[i]:= 0;
  FAtoms.GPIO.crc16:= CRC16CCITT(@FAtoms.GPIO, SizeOf(FAtoms.GPIO)-SizeOf(FAtoms.GPIO.crc16));
  // Device tree blob ignored
  FAtoms.DeviceTreeBlob.aType:= 0;
  // update total length
  FAtoms.Header.eepLen:= SizeOf(FAtoms.Header)
                    + SizeOf(FAtoms.Vendor.aType) + SizeOf(FAtoms.Vendor.count)
                     + SizeOf(FAtoms.Vendor.dLen) + FAtoms.Vendor.dLen
                    + SizeOf(FAtoms.GPIO);
end;

procedure THat_eeprom_ctlr.AtomsToBuffer;
// Set the buffer from Atoms
var i: UInt16;
    Addr, VendorFixedLen, BlocLen: UInt32;
begin
  Addr:= 0;
  if FAtoms.Header.numAtoms > 0 then
  begin
    //Header
    Move(FAtoms.Header, FeBuf[Addr], SizeOf(FAtoms.Header));
    Addr:= Addr + SizeOf(FAtoms.Header);
    // Vendor
    if FAtoms.Vendor.aType = 1 then
    begin
      VendorFixedLen:= AtomStartLen + SizeOf(FAtoms.Vendor.UUID)
                    + SizeOf(FAtoms.Vendor.pid) + SizeOf(FAtoms.Vendor.pid)
                    + SizeOf(FAtoms.Vendor.vslen) + SizeOf(FAtoms.Vendor.vplen);
      Move(FAtoms.Vendor, FeBuf[Addr], VendorFixedLen);
      Addr:= Addr + VendorFixedLen;
      Move(FAtoms.Vendor.vstr[1], FeBuf[Addr], FAtoms.Vendor.vslen);
      Addr:= Addr + FAtoms.Vendor.vslen;
      Move(FAtoms.Vendor.pstr[1], FeBuf[Addr], FAtoms.Vendor.vplen);
      Addr:= Addr + FAtoms.Vendor.vplen;
      Move(FAtoms.Vendor.crc16, FeBuf[Addr], SizeOf(FAtoms.Vendor.crc16));
      Addr:= Addr + SizeOf(FAtoms.Vendor.crc16);
    end;
    // GPIO
    if FAtoms.GPIO.aType = 2 then
    begin
      Move(FAtoms.GPIO, FeBuf[Addr], SizeOf(FAtoms.GPIO));
      Addr:= Addr + SizeOf(FAtoms.GPIO);
    end;
    // Tree
    if FAtoms.DeviceTreeBlob.aType = 3 then
    begin
      BlocLen:= AtomStartLen + FAtoms.DeviceTreeBlob.dLen;
      Move(FAtoms.DeviceTreeBlob, FeBuf[Addr], BlocLen);
      Addr:= Addr + BlocLen;
    end;
    // Customs
    if Length(FAtoms.Customs) > 0 then
    for i:= 0 to Length(FAtoms.Customs)-1 do
    begin
      BlocLen:= AtomStartLen + FAtoms.Customs[i].dLen;
      Move(FAtoms.Customs[i], FeBuf[Addr], BlocLen);
      Addr:= Addr + BlocLen;
    end;
  end;
  // blank remaining
  for i:= Addr to Length(FeBuf)-1 do FeBuf[i]:= $FF;
end;

function THat_eeprom_ctlr.AddCustom(const AData; ALen: UInt16): UInt16;
var aIx, aNo: Integer;
begin
  aNo:= FAtoms.Header.numAtoms;
  aIx:= Length(FAtoms.Customs);
  // set Custom
  SetLength(FAtoms.Customs, aIx+1);
  with FAtoms.Customs[aIx] do
  begin
    aType:= 4;
    dLen:= ALen + SizeOf(crc16);
    count:= aNo;
    dataOffset:= FAtoms.Header.eepLen + AtomStartLen;
    SetLength(data, ALen);
    Move(AData, data[0], ALen);
    crc16:= CRC16CCITT(@FeBuf[SizeOf(FAtoms.Header)], AtomStartLen + dLen - SizeOf(crc16));
  end;
  // update Header
  FAtoms.Header.numAtoms:= aNo+1;
  FAtoms.Header.eepLen:= FAtoms.Customs[aIx].dataOffset + FAtoms.Customs[aIx].dLen;
  // return atom no
  Result:= aNo;
end;

procedure THat_eeprom_ctlr.SaveToCustom(AtomIx: UInt16; const AData; ALen: UInt16);
var Offset: UInt16;
begin
  if not FInitialized then Initialize;
  // Controls
  if AtomIx >= Length(FAtoms.Customs)
  then begin
    FLastError:= 'Custom atom not initialized';
    Exit;
  end;
  if ALen <> FAtoms.Customs[AtomIx].dLen - SizeOf(TheaCustom.crc16)
  then begin
    FLastError:= 'Custom data size error';
    Exit;
  end;
  // move to atom and buffer
  Move(AData, FAtoms.Customs[AtomIx].data[0], ALen);
  Offset:= FAtoms.Customs[AtomIx].dataOffset;
  Move(AData, FeBuf[Offset], ALen);
  // reset crc16
  FAtoms.Customs[AtomIx].crc16:= CRC16CCITT(@FeBuf[Offset-AtomStartLen],
         AtomStartLen + FAtoms.Customs[AtomIx].dLen - SizeOf(TheaCustom.crc16));
  Move(FAtoms.Customs[AtomIx].crc16, FeBuf[Offset+ALen], SizeOf(TheaCustom.crc16));
  // save to eeprom
  WriteArray(Offset, FeBuf, ALen+SizeOf(TheaCustom.crc16));

end;



procedure THat_eeprom_ctlr.GetFromCustom(AtomIx: UInt16; var AData;
  ALen: UInt16);
begin
  if not FInitialized then Initialize;
  // Controls
  if AtomIx >= Length(FAtoms.Customs)
  then begin
    FLastError:= 'Custom atom not initialized';
    Exit;
  end;
  if ALen <> FAtoms.Customs[AtomIx].dLen - SizeOf(TheaCustom.crc16)
  then begin
    FLastError:= 'Custom data size error';
    Exit;
  end;
  // read from custom atom
  Move(FAtoms.Customs[AtomIx].data[0], AData, ALen);
end;

function THat_eeprom_ctlr.CRC16CCITT(PBytes: PByte; ALen:UInt16): UInt16;
var
  PtrB: PByte;
  crc, i: UInt16;
  b: byte;
begin
  crc:= $FFFF;
  PtrB:= PBytes;
  for i:= 0 to ALen-1 do
  begin
    b:= byte(PtrB^);
    b:= byte(crc shr 8) xor b;
    b:= b xor (b shr 4);
    crc:= (crc shl 8) xor (UInt16(b) shl 12) xor (UInt16(b) shl 5) xor UInt16(b) ;
    Inc(PtrB);
  end;
  Result:= crc;
end;


end.
